# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4 nu

import collections
import re
import urllib.parse

import bs4

from .shared import GlobalMixin, logger
from .utils import normalize_ident, rebuild_uri

seldef = collections.namedtuple(
    "SelectorDef",
    [
        "selector",
        "for_category",
        "target_is_parent",
        "clear_parent_previous_sibling",
    ],
)


def remove_link_for_exclusion(link, selector_def):
    """remove a link or its parent based on selector_def"""
    if selector_def.clear_parent_previous_sibling:
        link.parent.previous_sibling.replace_with("")

    if selector_def.target_is_parent:
        link.parent.decompose()
    else:
        link.decompose()


def is_in_code(elem):
    """whether this node is inside a <code /> one

    <code/> blocks are used to share code and are thus usually not rewritten"""
    for parent in elem.parents:
        if parent.name in ("code", "pre"):
            return True
    return False


def is_child_of(elem, selector, soup):
    """whether element is a child of element mathcing selector"""
    target = soup.select(selector)
    return target and target[0] in elem.parents


def has_parent_tagged(elem, name):
    """whether element has a parent which such tag name"""
    for parent in elem.parents:
        if parent.name == name:
            return parent


class Rewriter(GlobalMixin):
    def __init__(self):
        self.domain_re = re.compile(rf"https?://{self.conf.domain}(?P<path>/.+)")

        # sets of articles and categories that should not be included
        # filled by --exclude option

    def rewrite(self, content: str, to_root: str = "", unwrap: bool = False):
        if not content:
            return ""

        try:
            soup = bs4.BeautifulSoup(content, "lxml")
        except Exception as exc:
            logger.error(f"Unable to init soup for {content}: {exc}")
            return content
        if not soup:
            return ""

        for attr in ("body", "html"):
            getattr(soup, attr).unwrap()

        self.rewrite_links(soup, to_root)
        self.rewrite_links_for_excludes(soup, to_root)
        self.remove_empty_sections(soup, to_root)

        self.rewrite_pictures(soup, to_root)

        self.rewrite_images(soup, to_root)

        return str(soup)

    def rewrite_links(self, soup, to_root):
        # rewrite links targets
        for link in soup.find_all("a", href=True):

            # don't bother empty href=""
            if not link.get("href", "").strip():
                # remove link to ""
                del link["href"]
                continue

            # Ignore reference links on the page
            if link["href"].startswith("#"):
                continue

            # skip links inside <code /> nodes
            if is_in_code(link):
                continue

            link["href"] = link["href"].strip()

            is_relative = link["href"][0] in ("/", ".") or not link["href"].startswith(
                "http"
            )

            if not is_relative:
                match = self.domain_re.match(link["href"])
                if match:
                    is_relative = True
                    # make the link relative and remove / so it's Zim compat
                    link["href"] = match.groupdict().get("path")[1:]

            # rewrite relative links to match our in-zim URIs
            if is_relative:
                # might be a relative link for which we don't offer an offline
                # version. ex: /help/*
                if not self.rewrite_relative_link(link, to_root):
                    continue

            # link is not relative, apply rules
            self.rewrite_external_link(link)

    def rewrite_external_link(self, link):
        link["class"] = " ".join(link.get("class", []) + ["external-link"])
        if self.conf.without_external_links:
            del link["href"]

    def rewrite_relative_link(self, link: bs4.element.Tag, to_root: str) -> bool:
        """whether to consider link as non-relative because we failed to rewrite"""

        # link to root (/)
        if link["href"] == "":
            link["href"] = to_root
            return

        try:
            uri = urllib.parse.urlparse(link["href"])
            # normalize path as if from root.
            # any folder-walking link is considered to be targetting root
            uri_path = re.sub(r"^(\.\.?/)+", "", uri.path)
            uri_path = re.sub(r"^/", "", uri_path)
        except Exception as exc:
            logger.error(f"Failed to parse link target {link['href']}: {exc}")
            # consider this external
            return True

        # Normalize with to_root
        link["href"] = rebuild_uri(
            uri=uri,
            path=f"{to_root}{uri_path}",
            failsafe=True,
        ).geturl()

    def rewrite_links_for_excludes(self, soup, to_root):
        # Stop execution of this function if program is full mode.
        if self.conf.full_mode:
            return

        article_re = re.compile(r"^" + to_root + r"(?P<path>.+)")
        category_re = re.compile(
            r"^" + to_root + self.metadata["category_prefix"] + r":(?P<path>.+)"
        )

        selectors = [
            # related articles in article page
            seldef("#relatedwikihows > a.related-wh[href]", False, False, False),
            # related articles in sidebar
            seldef(".related_articles > a.related-wh[href]", False, False, False),
            # link to article in category page
            seldef("#cat_all div.responsive_thumb > a[href]", False, True, False),
            # sub category link in category page
            seldef("#subcats > ul a[href]", True, True, False),
            # categorlyListing thumbnail link (english)
            seldef(".cat_container a[href]", True, True, False),
            # categorylisting link to category
            seldef("#catlist_container #catlist a[href]", True, True, False),
            # top breadcrumb in article page
            seldef(".breadcrumbs a[href]", True, True, False),
            # top breadcrumb in article page
            seldef("#footer_crumbs .breadcrumbs a[href]", True, True, False),
            # list of categories article is in  - in About section of article page
            seldef(".sp_box.sp_fullbox a[href]", True, True, True),
            seldef(".cat_grid .responsive_thumb a[href]", True, True, True),
        ]

        for sdef in selectors:
            for link in soup.select(sdef.selector):
                try:
                    href = normalize_ident(link.attrs["href"])
                except TypeError:
                    logger.debug(f"Invalid link during rewrite: {link}")
                    continue
                if sdef.for_category:
                    match = category_re.match(href)
                    match_path = match.groupdict().get("path") if match else None
                    if match_path and match_path not in self.expected_categories:
                        remove_link_for_exclusion(link, sdef)
                else:
                    match = article_re.match(href)
                    match_path = match.groupdict().get("path") if match else None
                    if match_path and match_path not in self.expected_articles:
                        remove_link_for_exclusion(link, sdef)

    def remove_empty_sections(self, soup, to_root):
        """remove sections left empty by links removal; if any"""
        removed_relateds = False
        for section_sel in (".section.relatedwikihows", ".sidebox.related_articles"):
            for section in soup.select(section_sel):
                if not section.select("a"):
                    removed_relateds = True
                    section.decompose()

        if removed_relateds:
            _ = [item.decompose() for item in soup.select("#rwh_toc")]

    def rewrite_images(self, soup, to_root):
        for img in soup.find_all("img"):
            if img.attrs.get("onload"):
                del img.attrs["onload"]

            if not img.get("src") and not img.get("data-src"):
                continue

            if img.get("data-src") and not img.get("src"):
                img["src"] = img["data-src"]
                del img["data-src"]
                try:
                    del img["data-src-nowebp"]
                except KeyError:
                    pass

            # skip links inside <code /> nodes
            if is_in_code(img):
                continue

            # will introduce Webp Polyfill later
            # img["onerror"] = "onImageLoadingError(this);"
            path = self.imager.defer(img["src"])
            if path is None:
                del img["src"]
            else:
                img["src"] = f"{to_root}{path}"

    def rewrite_pictures(self, soup, to_root):
        """remove all elements (source mostly) inside picture except img

        pictures are used where there are webp version and the img is used as
        fallback."""
        for picture in soup.find_all("picture"):
            # there's no fallback img, we can't remove sources
            if not picture.find("img"):
                continue

            for child in picture.contents:
                if isinstance(child, bs4.element.Tag) and child.name != "img":
                    child.decompose()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4 nu

import collections
import io
import re
import urllib.parse
import zlib
from typing import Iterable, List, Tuple, Union

import bs4
import cssbeautifier
import requests
from kiwixstorage import KiwixStorage
from pif import get_public_ip
from tld import get_fld
from zimscraperlib.download import _get_retry_adapter, stream_file

from .shared import Global, logger

nlink = collections.namedtuple("Link", ("path", "name", "title"))


def to_path(url: str) -> str:
    """Path-part of an URL, without leading slash"""
    return re.sub(r"^/", "", urllib.parse.urlparse(url).path)


def get_url(path: str, **params) -> str:
    """url-encoded in-source website url for a path"""
    params_str = f"?{urllib.parse.urlencode(params)}" if params else ""
    return f"{Global.conf.main_url.geturl()}{urllib.parse.quote(path)}{params_str}"


def get_url_raw(path: str):
    """in-source website url for a path, untainted"""
    return f"{Global.conf.main_url.geturl()}{path}"


def to_url(value: str) -> str:
    """resolved potentially relative url from in-source link"""
    return value if value.startswith("http") else get_url_raw(value)


def to_rel(url: str) -> Union[None, str]:
    """path from URL if on our main domain, else None"""
    uri = urllib.parse.urlparse(url)
    if uri.netloc != Global.conf.domain:
        return None
    return uri.path


def no_leading_slash(text: str) -> str:
    """text with leading slash removed if present"""
    return re.sub(r"^/", "", text)


def no_trailing_slash(text: str) -> str:
    """text with trailing slash removed if present"""
    return re.sub(r"/$", "", text)


def only_path_of(url: str):
    """normalized path part of an url"""
    return normalize_ident(urllib.parse.urlparse(url).path)


def fetch(path: str, **params) -> str:
    """(source text, actual_paths) of a path from source website

    actual_paths is amn ordered list of paths that were traversed to get to content.
    Without redirection, it should be a single path, equal to request
    Final, target path is always last"""
    session = requests.Session()
    session.mount("http", _get_retry_adapter(5))  # tied to http and https
    resp = session.get(get_url(path, **params), params=params)
    resp.raise_for_status()

    # we have params meaning we requested a page (?pg=xxx)
    # assumption: this must be a category page (so on same domain)
    # we thus need to use redirection target (which lost param) with params
    if params and resp.history:
        return fetch(only_path_of(resp.url), **params)
    return resp.text, [
        no_leading_slash(only_path_of(r.url)) for r in resp.history + [resp]
    ]


def get_soup_of(text: str, unwrap: bool = False):
    """an lxml soup of an HTML string"""
    soup = bs4.BeautifulSoup(text, "lxml")
    if unwrap:
        for elem in ("body", "html"):
            getattr(soup, elem).unwrap()
    return soup


def get_footer_crumbs_from(soup: bs4.element.Tag) -> List[Tuple[str, str, str]]:
    """List of (url, name and title) of footer breadcrumbs"""

    crumbs = []
    for link in soup.select("#footer_crumbs ul li a[href]"):
        if not link.attrs.get("href"):
            continue
        # might or might not be a Category link
        try:
            cat_ident = cat_ident_for(link.attrs["href"])
        except Exception:
            cat_ident = None
        if cat_ident is None or cat_ident in Global.expected_categories:
            crumbs.append(
                nlink(link.attrs["href"][1:], link.string, link.attrs.get("title"))
            )
    return crumbs


def get_footer_links_from(soup: bs4.element.Tag) -> List[Tuple[str, str, str]]:
    """list of namedtuple(path, name, title) of footer links"""
    links = []

    fld = get_fld(Global.conf.main_url.geturl())

    # Skip some links with no offline value
    for link in soup.select("#footer_links ul li a"):
        if link.attrs.get("href") in (
            "#",
            "https://blog.wikihow.com/",
            "/wikiHow:Jobs",
            "https://www.wikihow.com/wikiHow:Contribute",
        ):
            continue

        path = None
        if link.attrs.get("href"):
            url = urllib.parse.urlparse(to_url(link.attrs["href"]))

            # skip external URLs (if any)
            if get_fld(url.geturl()) != fld:
                continue
            path = normalize_ident(url.path)[1:]

        links.append(nlink(path, link.string, link.attrs.get("title")))
    return links


def get_soup(path: str, **params) -> bs4.BeautifulSoup:
    """an lxml soup of a path on source website"""
    content, paths = fetch(path, **params)
    return get_soup_of(content), paths


def soup_link_finder(elem: bs4.element.Tag) -> bool:
    """bs4's find_all-friendly selector for linked styles in wikiHow"""
    return (
        elem.name == "link"
        and elem.attrs.get("href")
        and (elem.attrs.get("as") == "style" or elem.attrs.get("rel") == "stylesheet")
    )


def get_digest(url: str) -> str:
    """simple digest of an url for mapping purpose"""
    return str(zlib.adler32(url.encode("UTF-8")))


def cat_ident_for(href: str) -> str:
    """decoded name of a category from a link target"""
    return normalize_ident(href).split(":", 1)[1]


def fix_pagination_links(soup: bs4.element.Tag):
    """Replace ?pg= to _pg= in pagination links"""
    for a in soup.select("#large_pagination a[href]"):
        a["href"] = a["href"].replace("?pg=", "_pg=")


def get_categorylisting_url():
    return normalize_ident(
        urllib.parse.urlparse(requests.get(to_url("/Special:CategoryListing")).url).path
    )[1:]


def get_youtube_id_from(url: str) -> str:
    """Youtube video Id from a youtube URL"""
    uri = urllib.parse.urlparse(url)
    if uri.path.startswith("/embed/"):
        m = re.match(r"^/embed/(?P<id>[^/]+)", uri.path)
        if m:
            return m.groupdict().get("id")
    if uri.path.startswith("/watch"):
        return urllib.parse.parse_qs(uri.query).get("v", [None]).pop()


def normalize_youtube_url(url: str) -> str:
    """harmonize youtube-URL to use a single (public viewing) format

    format: https://www.youtube.com/watch?v=C1vI8k-JEsQ"""

    yid = get_youtube_id_from(url)
    if yid:
        return f"https://www.youtube.com/watch?v={yid}"
    return url


def normalize_ident(ident: str) -> str:
    """URL-decoded category identifier"""
    return urllib.parse.unquote(ident)


def article_ident_for(href: str) -> str:
    """decoded name of an article from a link target"""
    return normalize_ident(to_rel(href))[1:]


def parse_css(style: str) -> Tuple[str, List[Tuple[str, str]]]:
    """(css, resources) of transformed CSS string and resources list

    reads a CSS string and returns it transformed
    with url() replaced by offlinable path.

    resources list is list of tuples, each containing the URL to get data from
    and the path to store it at.
    ex: ("http://goto.img/hello.png", "img/hello.png")"""

    output = ""
    resources = []

    def write(line):
        nonlocal output
        output += line + "\n"

    pattern = "url("
    for line in cssbeautifier.beautify(style).split("\n"):
        if pattern not in line:
            write(line)
            continue

        start = line.index(pattern) + len(pattern)
        end = line.index(")")

        # check whether it's quoted or not
        if line[start + 1] in ("'", '"'):
            start += 1
            end -= 1

        url = line[start:end]

        if url.startswith("data:"):
            write(line)
            continue

        path = f"assets/{get_digest(url)}"
        resources.append((url, path))
        # resources are added on same level (assets/xxx) as css itself
        write(line[0:start] + "../" + path + line[end:])

    return output, resources


def first(*args: Iterable[object]) -> object:
    """first non-None value from *args ; fallback to empty string"""
    return next((item for item in args if item is not None), "")


def rebuild_uri(
    uri: urllib.parse.ParseResult,
    scheme: str = None,
    username: str = None,
    password: str = None,
    hostname: str = None,
    port: Union[str, int] = None,
    path: str = None,
    params: str = None,
    query: str = None,
    fragment: str = None,
    failsafe: bool = False,
) -> urllib.parse.ParseResult:
    """new named tuple from uri with requested part updated"""
    try:
        username = first(username, uri.username, "")
        password = first(password, uri.password, "")
        hostname = first(hostname, uri.hostname, "")
        port = first(port, uri.port, "")
        netloc = (
            f"{username}{':' if password else ''}{password}"
            f"{'@' if username or password else ''}{hostname}"
            f"{':' if port else ''}{port}"
        )
        return urllib.parse.urlparse(
            urllib.parse.urlunparse(
                (
                    first(scheme, uri.scheme),
                    netloc,
                    first(path, uri.path),
                    first(params, uri.params),
                    first(query, uri.query),
                    first(fragment, uri.fragment),
                )
            )
        )
    except Exception as exc:
        if failsafe:
            logger.error(
                f"Failed to rebuild "  # lgtm [py/clear-text-logging-sensitive-data]
                f"URI {uri} with {scheme=} {username=} {password=} "
                f"{hostname=} {port=} {path=} "
                f"{params=} {query=} {fragment=} - {exc}"
            )
            return uri
        raise exc


def get_version_ident_for(url: str) -> str:
    """~version~ of the URL data to use for comparisons. Built from headers"""
    try:
        resp = requests.head(url)
        headers = resp.headers
    except Exception:
        logger.warning(f"Unable to HEAD {url}")
        try:
            _, headers = stream_file(
                url=url,
                byte_stream=io.BytesIO(),
                block_size=1,
                only_first_block=True,
            )
        except Exception:
            logger.warning(f"Unable to query image at {url}")
            return

    for header in ("ETag", "Last-Modified", "Content-Length"):
        if headers.get(header):
            return headers.get(header)

    return "-1"


def setup_s3_and_check_credentials(s3_url_with_credentials):
    logger.info("testing S3 Optimization Cache credentials")
    s3_storage = KiwixStorage(s3_url_with_credentials)
    if not s3_storage.check_credentials(
        list_buckets=True, bucket=True, write=True, read=True, failsafe=True
    ):
        logger.error("S3 cache connection error testing permissions.")
        logger.error(f"  Server: {s3_storage.url.netloc}")
        logger.error(f"  Bucket: {s3_storage.bucket_name}")
        logger.error(f"  Key ID: {s3_storage.params.get('keyid')}")
        logger.error(f"  Public IP: {get_public_ip()}")
        raise ValueError("Unable to connect to Optimization Cache. Check its URL.")
    return s3_storage



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ai ts=4 sts=4 et sw=4 nu

import io
import pathlib
import re
import urllib.parse
from typing import Optional, Union

import youtube_dl
from kiwixstorage import KiwixStorage, NotFoundError
from tld import get_fld
from zimscraperlib.video.encoding import reencode
from zimscraperlib.video.presets import VideoWebmHigh, VideoWebmLow

from .constants import VIDEOS_ENCODER_VERSION
from .shared import Global
from .utils import (
    get_digest,
    get_version_ident_for,
    get_youtube_id_from,
    normalize_ident,
    normalize_youtube_url,
)

logger = Global.logger


class VideoGrabber:
    def __init__(self):
        self.aborted = False
        # list of source URLs that we've processed and added to ZIM
        self.handled = set()
        self.nb_requested = 0
        self.nb_done = 0

        Global.video_executor.start()

    @staticmethod
    def youtube_poster_url(url: str):
        yid = get_youtube_id_from(url)
        if yid:
            return f"http://img.youtube.com/vi_webp/{yid}/sddefault.webp"

    @property
    def videos_dir(self):
        return Global.conf.build_dir.joinpath("videos")

    def abort(self):
        """request videograbber to cancel processing of requests"""
        self.aborted = True

    def get_url(self, url: str) -> urllib.parse.ParseResult:
        """normalized URL from requested one"""
        is_youtube = get_fld(url) == "youtube.com"
        if is_youtube:
            url = normalize_youtube_url(url)
        return urllib.parse.urlparse(url), is_youtube

    def get_video_fpath(self, url: str, is_youtube: bool) -> io.BytesIO:
        """fpath of downloadd and processed video from url"""

        # we only support WebM
        audext, vidext = ("webm", "webm")
        preset = VideoWebmLow() if Global.conf.low_quality else VideoWebmHigh()

        digest = get_digest(url)

        # download video using youtube-dl.
        # used for both youtube videos and standard (.mp4) urls
        options = {
            "cachedir": self.videos_dir,
            "writethumbnail": False,
            "writesubtitles": False,
            "allsubtitles": False,
            "writeautomaticsub": False,
            "subtitlesformat": "vtt",
            "keepvideo": False,
            "ignoreerrors": False,
            "retries": 20,
            "fragment-retries": 50,
            "skip-unavailable-fragments": True,
            "outtmpl": str(self.videos_dir.joinpath(digest + ".%(ext)s")),
            "preferredcodec": Global.conf.video_format,
            "format": f"best[ext={vidext}]/bestvideo[ext={vidext}]+"
            f"bestaudio[ext={audext}]/best",
            "y2z_videos_dir": self.videos_dir,
            "nocheckcertificate": True,
        }
        with youtube_dl.YoutubeDL(options) as ydl:
            ydl.download([url])

        # find downloaded videos from youtube-dl output dir
        # extension is unknown at this stage. filename should respect
        # the `outtmpl` format we gave to youtube-dl though.
        files = [
            p
            for p in self.videos_dir.iterdir()
            if p.stem == digest and p.suffix not in (".jpg", ".webp")
        ]
        if len(files) == 0:
            logger.error(f"Video file missing in {self.videos_dir} for {url}")
            logger.debug(list(self.videos_dir.iterdir()))
            raise FileNotFoundError(f"Missing video file in {self.videos_dir}")
        if len(files) > 1:
            logger.warning(
                f"Multiple video file candidates for {url} in {self.videos_dir}."
                f" Picking {files[0]} out of {files}"
            )
        src_path = files[0]

        # skip reencoding if youtube-dl gave us a WebM file and we don't need low-Q
        if (
            not Global.conf.low_quality
            and src_path.suffix[1:] == Global.conf.video_format
        ):
            return src_path

        # reencode if format is different or quality must be reduced
        dst_path = src_path.with_name(f"{src_path.stem}-2.{Global.conf.video_format}")
        reencode(
            src_path,
            dst_path,
            preset.to_ffmpeg_args(),
            delete_src=False,
            failsafe=False,
        )

        return dst_path

    def get_s3_key_for(self, url: str) -> str:
        """S3 key to use for that url"""
        return re.sub(r"^(https?)://", r"\1/", url)

    def get_path_for(self, url: urllib.parse.ParseResult) -> str:
        suffix = ".webm"
        digest = get_digest(url.geturl())
        return f"videos/{digest}-{normalize_ident(pathlib.Path(url.path).stem)}{suffix}"

    def defer(
        self,
        url: str,
        path: Optional[str] = None,
    ) -> Union[str, None]:
        """request full processing of url, returning in-zim path immediately"""

        # find actual URL should it be from a provider
        try:
            url, is_youtube = self.get_url(url)
        except Exception:
            logger.warning(f"Can't parse video URL `{url}`. Skipping")
            return

        if url.scheme not in ("http", "https"):
            logger.warning(f"Not supporting video URL `{url.geturl()}`. Skipping")
            return

        digest = get_digest(url.geturl())
        path = self.get_path_for(url) if path is None else path

        # skip processing if we already processed it or have it in pipe
        if digest in self.handled:
            logger.debug(f"URL `{url.geturl()}` already processed.")
            return path

        # record that we are processing this one
        self.handled.add(digest)
        self.nb_requested += 1

        Global.video_executor.submit(
            self.process_video,
            url=url,
            is_youtube=is_youtube,
            path=path,
            dont_release=True,
        )

        return path

    def once_done(self):
        self.nb_done += 1
        logger.debug(f"Videos {self.nb_done}/{self.nb_requested}")

    def process_video(self, url: str, is_youtube: bool, path) -> str:
        """download video from url or S3 and add to ZIM at path. Upload if req."""

        if self.aborted:
            return

        # just download, optimize and add to ZIM if not using S3
        if not Global.conf.s3_url:
            with Global.lock:
                Global.creator.add_item_for(
                    path=path,
                    fpath=self.get_video_fpath(url.geturl(), is_youtube),
                    delete_fpath=True,
                    mimetype="video/webm",
                    callback=self.once_done,
                )
            return path

        # we are using S3 cache
        if is_youtube:
            ident = "1"
        else:
            ident = get_version_ident_for(url.geturl())
            if ident is None:
                logger.error(f"Unable to query {url.geturl()}. Skipping")
                return path

        key = self.get_s3_key_for(url.geturl())
        s3_storage = KiwixStorage(Global.conf.s3_url)
        meta = {"ident": ident, "encoder_version": str(VIDEOS_ENCODER_VERSION)}

        download_failed = False  # useful to trigger reupload or not
        try:
            logger.debug(f"Attempting download of S3::{key} into ZIM::{path}")
            fileobj = io.BytesIO()
            s3_storage.download_matching_fileobj(key, fileobj, meta=meta)
        except NotFoundError:
            # don't have it, not a donwload error. we'll upload after processing
            pass
        except Exception as exc:
            logger.error(f"failed to download {key} from cache: {exc}")
            logger.exception(exc)
            download_failed = True
        else:
            with Global.lock:
                Global.creator.add_item_for(
                    path=path,
                    content=fileobj.getvalue(),
                    mimetype="video/webm",
                    callback=self.once_done,
                )
            return path

        # we're using S3 but don't have it or failed to download
        try:
            fpath = self.get_video_fpath(url.geturl(), is_youtube)
        except Exception as exc:
            logger.error(f"Failed to download/convert/optim source  at {url.geturl()}")
            logger.exception(exc)
            return path

        with Global.lock:
            Global.creator.add_item_for(
                path=path,
                fpath=fpath,
                delete_fpath=True,
                mimetype="video/webm",
                callback=self.once_done,
            )

        # only upload it if we didn't have it in cache
        if not download_failed:
            logger.debug(f"Uploading {url.geturl()} to S3::{key} with {meta}")
            try:
                s3_storage.upload_file(fpath=fpath, key=key, meta=meta)
            except Exception as exc:
                logger.error(f"{key} failed to upload to cache: {exc}")

        return path



# -*- coding: utf-8 -*-

import argparse
import os
import sys

from .constants import NAME, SCRAPER, URLS
from .shared import Global, logger


def main():
    parser = argparse.ArgumentParser(
        prog=NAME,
        description="Scraper to create ZIM files wikihow articles",
    )

    parser.add_argument(
        "--language",
        choices=URLS.keys(),
        required=True,
        help="wikiHow website to build from",
        dest="lang_code",
    )

    parser.add_argument(
        "--output",
        help="Output folder for ZIM file",
        default="/output",
        dest="_output_dir",
    )

    parser.add_argument(
        "--name",
        help="ZIM name. Used as identifier and filename (date will be appended). "
        "Constructed from language if not supplied",
    )

    parser.add_argument(
        "--title", help="Custom title for your ZIM. Wikihow homepage title otherwise"
    )

    parser.add_argument(
        "--description",
        help="Custom description for your ZIM. "
        "Wikihow homepage description (meta) otherwise",
    )

    parser.add_argument(
        "--icon",
        help="Custom icon for your ZIM (path or URL). " "wikiHow square logo otherwise",
    )

    parser.add_argument(
        "--creator",
        help="Name of content creator. “wikiHow” otherwise",
        dest="author",
    )

    parser.add_argument(
        "--publisher", help="Custom publisher name (ZIM metadata). “openZIM” otherwise"
    )

    parser.add_argument(
        "--tag",
        help="Add tag to the ZIM file. "
        "_category:wikihow and wikihow added automatically. Use --tag several times or "
        "separate with `;`",
        default=[],
        action="append",
    )

    parser.add_argument(
        "--zim-file",
        help="ZIM file name (based on --name if not provided)",
        dest="fname",
    )

    parser.add_argument(
        "--category",
        help="Only scrape this category (can be specified multiple times). "
        "Use URL-ID of the Category "
        "(after the colon `:` in the URL). "
        "Add a slash after Category to request it without recursion",
        dest="categories",
        action="append",
    )

    parser.add_argument(
        "--low-quality",
        help="Use lower-quality, smaller file-size video encode",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "--without-videos",
        help="Don't include the video blocks (Youtube hosted). Most are copyrighted",
        action="store_true",
        default=False,
        dest="without_videos",
    )

    parser.add_argument(
        "--without-external-links",
        help="Don't include external links",
        action="store_true",
        default=False,
        dest="without_external_links",
    )

    parser.add_argument(
        "--exclude",
        help="Path or URL to a text file listing Article ID or `Category:` prefixed "
        "Category IDs to exclude from the scrape. Lines starting with # are ignored",
        dest="exclude",
    )

    parser.add_argument(
        "--only",
        help="Path or URL to a text file listing Article IDs to include in the scrape. "
        "This filters out every other article. Lines starting with # are ignored",
        dest="only",
    )

    parser.add_argument(
        "--optimization-cache",
        help="URL with credentials to S3 for using as optimization cache",
        dest="s3_url_with_credentials",
    )

    parser.add_argument(
        "--debug", help="Enable verbose output", action="store_true", default=False
    )

    parser.add_argument(
        "--tmp-dir",
        help="Path to create temp folder in. Used for building ZIM file. "
        "Videos are stored and re-encoded there.",
        default=os.getenv("TMPDIR", "."),
        dest="_tmp_dir",
    )

    parser.add_argument(
        "--keep",
        help="Don't remove build folder on start (for debug/devel)",
        default=False,
        action="store_true",
        dest="keep_build_dir",
    )

    parser.add_argument(
        "--build-in-tmp",
        help="Use --tmp-dir value as workdir. Otherwise, a unique sub-folder "
        "is created inside it. Useful to reuse downloaded files (debug/devel)",
        default=False,
        action="store_true",
        dest="build_dir_is_tmp_dir",
    )

    parser.add_argument(
        "--delay",
        help="Add this delay (seconds) before each request to please "
        "wikiHow servers. Can be fractions. Defaults to 0: no delay",
        type=float,
    )

    parser.add_argument(
        "--api-delay",
        help="Add this delay (seconds) before each API query (!= calls) to please "
        "wikiHow servers. Can be fractions. Defaults to 0: no delay",
        type=float,
    )

    parser.add_argument(
        "--skip-dom-check",
        help="[dev] Don't perform DOM Integrity Checks on start",
        default=False,
        action="store_true",
        dest="skip_dom_check",
    )

    parser.add_argument(
        "--skip-footer-links",
        help="[dev] Don't scrape footer links",
        default=False,
        action="store_true",
        dest="skip_footer_links",
    )

    parser.add_argument(
        "--single-article",
        help="[dev] Don't scrape categories, just that single article instead",
        dest="single_article",
    )

    parser.add_argument(
        "--stats-filename",
        help="Path to store the progress JSON file to.",
        dest="stats_filename",
    )

    parser.add_argument(
        "--version",
        help="Display scraper version and exit",
        action="version",
        version=SCRAPER,
    )

    args = parser.parse_args()
    Global.set_debug(args.debug)

    from .scraper import wikihow2zim

    try:
        scraper = wikihow2zim(**dict(args._get_kwargs()))
        sys.exit(scraper.run())
    except Exception as exc:
        logger.error(f"FAILED. An error occurred: {exc}")
        if args.debug:
            logger.exception(exc)
        raise SystemExit(1)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import pathlib
import sys


def main():
    # allows running it from source using python wikihow2zim
    sys.path = [str(pathlib.Path(__file__).parent.parent.resolve())] + sys.path

    from wikihow2zim.entrypoint import main as entry

    entry()


if __name__ == "__main__":
    main()
